import neuron_ml.core.public.load
import neuron_ml.core.public.export
import neuron_ml.core.public.train
import neuron_ml.core.public.clean
import neuron_ml.core.public.classify
import neuron_ml.core.public.graph
import neuron_ml.core.public.labels
import neuron_ml.core.public.model
import neuron_ml.core.public.image



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

__init__.py is the core of the module. It imports all
of the required components, so the module can be
imported directly, without needing to do multiple
import statements.

"""
import neuron_ml.core as core

load = core.public.load.load
train = core.public.train.train
export = core.public.export.export
clean = core.public.clean.clean
classify = core.public.classify.classify
graph = core.public.graph.graph
labels = core.public.labels.labels
model = core.public.model.model
image = core.public.image.image



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/public/labels.py defines the 'labels' method that
will be used to load labels from a file for the
'classify' method.

"""

def labels(model):
	if model[1] == "tensorflow":
		import tensorflow as tf

		label_file = model[2][1]

		label = []
		proto_as_ascii_lines = tf.gfile.GFile(label_file).readlines()
		for l in proto_as_ascii_lines:
			label.append(l.rstrip())
		print("[Neuron - Labels] " + str(len(label)) + " labels were loaded.")
		return label
	else:
		raise ValueError("[Neuron - Graph] ERROR: Only TensorFlow trained model are accepted.")
		return



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/public/export.py defines the 'export' method that
will be used to export the trained model.

"""


def export(model, path, method="default"):
	import shutil
	import subprocess
	import sys
	global_path = model[0]
	technology = model[1]
	outputed_files = model[2]
	if technology == "tensorflow" and len(outputed_files) == 2:
		if method == "default" and type(path) == list:
			for i in range(len(outputed_files)):
				shutil.copy2(outputed_files[i], path[i])
		elif method == "tflite":
			import neuron_ml.tools.subprocess_checker as valid
			if valid.command("toco"):
				command = [
					"toco",
					"--graph_def_file",
					outputed_files[0],
					"--output_file",
					path,
					"--input_arrays",
					"Placeholder",
					"--output_arrays",
					"final_result"
				]
				print(" ".join(command))
				process = subprocess.Popen(command, stdout=subprocess.PIPE)
				for line in iter(lambda: process.stdout.read(1), b''):
					sys.stdout.buffer.write(line)
			else:
				raise ValueError(
					"[Neuron - Export] ERROR: TOCO not found. Please install TensorFlow's TOCO cli.")
				return
		elif method == "coreml":
			import tfcoreml as tf_converter
			tf_converter.convert(tf_model_path=outputed_files[0],
								 mlmodel_path=path,
								 output_feature_names=[
									 'final_result:0'],
								 image_input_names='input',
								 class_labels=outputed_files[1],
								 red_bias=-1,
								 green_bias=-1,
								 blue_bias=-1,
								 image_scale=2.0 / 299.0
								 )
		else:
			raise ValueError(
				"[Neuron - Export] ERROR: Wrong arguments. See the wiki for help.")
			return
	elif technology == "createml" and method == "default":
		shutil.copy2(outputed_files[0], path)
	else:
		raise ValueError(
			"[Neuron - Export] ERROR: Wrong arguments. See the wiki for help.")
		return
	print("[Neuron - Export] All file where exported.")



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/public/image.py defines the 'image' method that
will be used to encode an image in order to classify
it.

"""

def image(file_name):
	import tensorflow as tf
	input_name = "file_reader"
	file_reader = tf.read_file(file_name, input_name)
	if file_name.endswith(".png"):
		return tf.image.decode_png(
			file_reader, channels=3, name="png_reader")
	elif file_name.endswith(".gif"):
		return tf.squeeze(
			tf.image.decode_gif(file_reader, name="gif_reader"))
	elif file_name.endswith(".bmp"):
		return tf.image.decode_bmp(file_reader, name="bmp_reader")
	else:
		return tf.image.decode_jpeg(
			file_reader, channels=3, name="jpeg_reader")



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/public/model.py defines the 'model' method that
will be used to generate a model object from pre
trained graphs

"""

def model(files, method="tensorflow"):
	return [0, method, files]



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/public/load.py defines the 'load' method that
will be used to prepare data for training.

"""

def load(dataset, method="tensorflow"):
	import os
	import tempfile
	import shutil
	import neuron_ml.core.data as data

	def createDir():
		return tempfile.mkdtemp()


	if method == "tensorflow":
		# Prepare and load for TensorFlow
		temp = createDir()
		data.tensorflow.organise_dataset(dataset, temp)
		return [temp, method]
	elif method == "createml":
		# Prepare and load data for CreateML
		temp = createDir()
		data.createml.organise_dataset(dataset, temp)
		return [temp, method]
	else:
		raise ValueError("[Neuron - Load] ERROR: Wrong method. You have 2 options: 'tensorflow' or 'createml'.")
		return



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

tools/subprocess_checker.py checks if a subprocess
command is available and if it's safe to spawn it.

"""

import os
import platform
import subprocess

def command(name):
	cmd = "where" if platform.system() == "Windows" else "which"
	exist = False
	try:
		subprocess.call([cmd, name])
		exist = True
	except:
		exist = False
	return exist



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/data/createml.py will transfer and convert
images to the desired destination (TEMP directory)
for CreateML usage.

"""

def organise_dataset(data, path):
	import os
	import sys
	import pandas as pd
	import shutil
	os.makedirs(path, exist_ok=True)
	folders = os.listdir(data)
	print("[Neuron - Load] Organising dataset by moving to dataset folder & converting to jpeg")
	for folder in folders:
		if folder != ".git" and os.path.isdir(data + '/' + folder):
			images = os.listdir(data + '/' + folder)
			count = len(images)
			n_test = int(round(count * 0.9)) # Will take 10% for tests. Max 10 images
			if n_test > 10:
				n_test = 10
			r_test = range(n_test)

			train = path + '/Train/' + folder
			test = path + '/Test/' + folder
			os.makedirs(train)
			os.makedirs(test)
			for i in range(count):
				p = folder + '/' +images[i]
				s = data + '/' + folder + '/' + images[i]
				if i in r_test: # Move to test
					d = path + '/Test/' + p
					shutil.copy2(s, d)
				else:
					d = path + '/Train/' + p
					shutil.copy2(s, d)
	print("[Neuron - Load] Dataset folders successfully created & copied all images in corresponding folders")



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/data/tensorflow.py will transfer and convert
images to the desired destination (TEMP directory).

"""

def toJPG(source, destination):
	from PIL import Image
	im = Image.open(source)
	rgb_im = im.convert('RGB')
	rgb_im.save(destination)


def organise_dataset(train, root_path):
	import os
	import sys
	import pandas as pd
	dataset_path = root_path
	train_data = train
	os.makedirs(dataset_path, exist_ok=True)
	folders = os.listdir(train_data)
	print("[Neuron - Load] Organising dataset by moving to dataset folder & converting to jpeg")
	for folder in folders:
		if folder != ".git" and os.path.isdir(train + '/' + folder):
			os.makedirs(dataset_path+'/'+folder, exist_ok=True)
			images = os.listdir(train + '/' + folder)
			for i in images:
				path = folder + '/' + i
				source = train_data + '/' + path
				destination = dataset_path + '/' + path
				# Moving files from source (train folder) to detination folder and convert to JPG
				toJPG(source, destination)


	print("[Neuron - Load] Dataset folders successfully created & copied all images in corresponding folders")



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/public/classify.py defines the 'classify' method that
will be used to classify any images using a given model.

"""


def classify(graph, labels, image):
	import numpy as np
	import tensorflow as tf
	import time
	start = time.time()
	def read_tensor_from_image_file(file_name,
									input_height=299,
									input_width=299,
									input_mean=0,
									input_std=255):
		input_name = "file_reader"
		output_name = "normalized"
		if type(file_name) != str:
			image_reader = file_name
		elif file_name.endswith(".png"):
			file_reader = tf.read_file(file_name, input_name)
			image_reader = tf.image.decode_png(
				file_reader, channels=3, name="png_reader")
		elif file_name.endswith(".gif"):
			file_reader = tf.read_file(file_name, input_name)
			image_reader = tf.squeeze(
				tf.image.decode_gif(file_reader, name="gif_reader"))
		elif file_name.endswith(".bmp"):
			file_reader = tf.read_file(file_name, input_name)
			image_reader = tf.image.decode_bmp(file_reader, name="bmp_reader")
		else:
			file_reader = tf.read_file(file_name, input_name)
			image_reader = tf.image.decode_jpeg(
				file_reader, channels=3, name="jpeg_reader")
		float_caster = tf.cast(image_reader, tf.float32)
		dims_expander = tf.expand_dims(float_caster, 0)
		resized = tf.image.resize_bilinear(
			dims_expander, [input_height, input_width])
		normalized = tf.divide(tf.subtract(resized, [input_mean]), [input_std])
		sess = tf.Session()
		result = sess.run(normalized)

		return result

	file_name = image
	input_height = 299
	input_width = 299
	input_mean = 0
	input_std = 255
	input_layer = "Placeholder"
	output_layer = "final_result"

	t = read_tensor_from_image_file(
		file_name,
		input_height=input_height,
		input_width=input_width,
		input_mean=input_mean,
		input_std=input_std)

	input_name = "import/" + input_layer
	output_name = "import/" + output_layer
	input_operation = graph.get_operation_by_name(input_name)
	output_operation = graph.get_operation_by_name(output_name)

	with tf.Session(graph=graph) as sess:
		results = sess.run(output_operation.outputs[0], {
			input_operation.outputs[0]: t
		})
	results = np.squeeze(results)

	top_k = results.argsort()[-5:][::-1]

	dictionnary = {
		"labels": labels,
		"top": "",
		"prob": 0.0,
		"results": {}
	}
	for i in top_k:
		dictionnary["results"][labels[i]] = results[i]
	max = np.amax(results)
	index = results.tolist().index(max)
	dictionnary["prob"] = results[index]
	dictionnary["top"] = labels[index]
	end = time.time()

	print("[Neuron - Classify] Classified as '" + dictionnary["top"] + "' in " + str(end - start) + "s.")
	return dictionnary



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/public/train.py defines the 'train' method that
will train the model.

"""

def train(loaded, steps=1000):
	import os
	import subprocess
	import sys
	import tempfile
	import shutil
	import inspect
	import neuron_ml.tools.subprocess_checker as valid
	current_file = inspect.getfile(inspect.currentframe())
	script_dir = os.path.abspath(current_file + '../../../../script')

	script_createml = script_dir + '/CreateML/trainer.swift'
	script_tensorflow = script_dir + '/TensorFlow/retrain.py'

	method = loaded[1]
	path = loaded[0]
	output = []
	print("[Neuron - Train] Start training the model using the '" + method + "' method.")
	if method == "tensorflow":
		command = [
			sys.executable,
			script_tensorflow,
			'--image_dir',
			path,
			'--output_graph',
			path + '/retrained_graph.pb',
			'--output_labels',
			path + '/retrained_labels.txt',
			'--how_many_training_steps',
			str(steps)
		]
		process = subprocess.Popen(command, stdout=subprocess.PIPE)
		for line in iter(process.stdout.readline, b''):
			sys.stdout.write(line)
		output.extend([path + '/retrained_graph.pb', path + '/retrained_labels.txt'])
	elif method == "createml":
		if valid.command("swift"):
			command = [
				"swift",
				script_createml,
				path,
				path + '/ExportedModel.mlmodel'
			]
			process = subprocess.Popen(command, stdout=subprocess.PIPE)
			for line in iter(lambda: process.stdout.read(1), b''):
				sys.stdout.buffer.write(line)
			output.extend([path + '/ExportedModel.mlmodel'])
		else:
			raise ValueError("[Neuron - Train] ERROR: Swift must be installed on your machine. Check https://swift.org for details.")
			return
	else:
		raise ValueError("[Neuron - Train] ERROR: Wrong method. You have 2 options: 'tensorflow' or 'createml'.")
		return
	print("[Neuron - Train] Training is done.")
	return [path, method, output]



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/public/clean.py defines the 'clean' method that
will be used to clean the temporary file that were
created.

"""

def clean(model):
	import shutil
	global_path = model[0]
	shutil.rmtree(global_path)



# -*- coding: utf-8 -*-
"""
-----------------------------------------------------
Copyright © Arthur Guiot 2018. All rights reserved
-----------------------------------------------------

/core/public/graph.py defines the 'graph' method that
will be used to load the graph from a file for the
'classify' method.

"""

def graph(model):
	if model[1] == "tensorflow":
		import tensorflow as tf
		model_file = model[2][0]
		graph = tf.Graph()
		graph_def = tf.GraphDef()

		with open(model_file, "rb") as f:
			graph_def.ParseFromString(f.read())
		with graph.as_default():
			tf.import_graph_def(graph_def)
		print("[Neuron - Graph] Graph was correctly loaded.")
		return graph
	else:
		raise ValueError("[Neuron - Graph] ERROR: Only TensorFlow trained model are accepted.")
		return
